"""
Defines boolean predicates that can be evaluated against a recipe.

Filters allow easy searching within the structure of a :class:`recipemd.data.Recipe`. They can be created from strings
with the :class:`FilterParser` or in code with the :class:`FilterBuilder` which can be imported as :data:`f` from this
module.

"""
import itertools
import re
import unicodedata
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import reduce, wraps
from typing import List, Iterable, Pattern, Type, Callable, Optional

from pyparsing import infixNotation, QuotedString, CaselessKeyword, opAssoc, ParserElement, Combine, \
    ParseResults, MatchFirst, Regex

from recipemd.data import Recipe

__all__ = [
    "FilterParser", "f", "FilterBuilder",
    "BooleanAndOperation", "BooleanOrOperation", "BooleanNotOperation", "BooleanXorOperation",
    "AnyFilterTerm", "TagFilterTerm", "IngredientFilterTerm", "UnitFilterTerm",
    "FuzzyFilterString", "ExactFilterString", "RegexFilterString",
]


def _value_error_to_not_implemented(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValueError:
            return NotImplemented

    return wrapper


@dataclass(frozen=True)
class _ASTElement(ABC):
    @classmethod
    @abstractmethod
    def _create_from_tokens(cls, toks):
        raise NotImplementedError

    def _to_filter_element(self, other) -> '_FilterElement':
        if isinstance(other, _FilterElement):
            return other
        if isinstance(other, _FilterString):
            return AnyFilterTerm(other)
        if isinstance(other, str):
            return AnyFilterTerm(FuzzyFilterString(other))
        if isinstance(other, Pattern):
            return AnyFilterTerm(RegexFilterString(other))
        raise ValueError()

    @_value_error_to_not_implemented
    def __or__(self, other):
        return BooleanOrOperation(operands=[self._to_filter_element(self), self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __and__(self, other):
        return BooleanAndOperation(operands=[self._to_filter_element(self), self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __xor__(self, other):
        return BooleanXorOperation(operands=[self._to_filter_element(self), self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __ror__(self, other):
        return BooleanOrOperation(operands=[self._to_filter_element(other), self._to_filter_element(self)])

    @_value_error_to_not_implemented
    def __rand__(self, other):
        return BooleanAndOperation(operands=[self._to_filter_element(other), self._to_filter_element(self)])

    @_value_error_to_not_implemented
    def __rxor__(self, other):
        return BooleanXorOperation(operands=[self._to_filter_element(other), self._to_filter_element(self)])

    @_value_error_to_not_implemented
    def __invert__(self):
        return BooleanNotOperation(operands=[self._to_filter_element(self)])


@dataclass(frozen=True)
class _FilterString(_ASTElement, ABC):
    @abstractmethod
    def contained_in(self, to_search: Iterable[str]) -> bool:
        """Checks if any of the elements in `to_search` match the filter string."""
        raise NotImplementedError


@dataclass(frozen=True)
class FuzzyFilterString(_FilterString):
    string: str

    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(string=toks["string"])

    def contained_in(self, to_search: Iterable[str]) -> bool:
        to_find_caseless = _normalize_str(self.string)
        return any(to_find_caseless in _normalize_str(el) for el in to_search)


@dataclass(frozen=True)
class ExactFilterString(_FilterString):
    string: str

    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(string=toks["string"])

    def contained_in(self, to_search: Iterable[str]) -> bool:
        to_find_caseless = _normalize_str(self.string)
        return any(to_find_caseless == _normalize_str(el) for el in to_search)


@dataclass(frozen=True)
class RegexFilterString(_FilterString):
    regex: Pattern

    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(regex=re.compile(toks[0]))

    def contained_in(self, to_search: Iterable[str]) -> bool:
        return any(self.regex.search(el) for el in to_search)


@dataclass(frozen=True)
class _FilterElement(_ASTElement, ABC):
    @abstractmethod
    def evaluate(self, recipe: Recipe) -> bool:
        """Evaluate filter against given recipe"""
        raise NotImplementedError


@dataclass(frozen=True)
class FilterTerm(_FilterElement, ABC):
    filter_string: _FilterString

    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(filter_string=toks["filter_string"])


@dataclass(frozen=True)
class TagFilterTerm(FilterTerm):
    def evaluate(self, recipe: Recipe) -> bool:
        return self.filter_string.contained_in(recipe.tags)


@dataclass(frozen=True)
class IngredientFilterTerm(FilterTerm):
    def evaluate(self, recipe: Recipe) -> bool:
        ingredient_names = (ingr.name for ingr in recipe.leaf_ingredients if ingr.name is not None)
        return self.filter_string.contained_in(ingredient_names)


@dataclass(frozen=True)
class UnitFilterTerm(FilterTerm):
    def evaluate(self, recipe: Recipe) -> bool:
        ingredient_units = (ingr.amount.unit for ingr in recipe.leaf_ingredients if ingr.amount is not None and ingr.amount.unit is not None)
        yield_units = (yield_.unit for yield_ in recipe.yields if yield_.unit is not None)
        return self.filter_string.contained_in(itertools.chain(ingredient_units, yield_units))


@dataclass(frozen=True)
class AnyFilterTerm(IngredientFilterTerm, TagFilterTerm, UnitFilterTerm):
    def evaluate(self, recipe: Recipe) -> bool:
        return IngredientFilterTerm.evaluate(self, recipe) or TagFilterTerm.evaluate(self, recipe) \
               or UnitFilterTerm.evaluate(self, recipe)


@dataclass(frozen=True)
class _BooleanOperation(_FilterElement, ABC):
    operands: List[_FilterElement]


@dataclass(frozen=True)
class _BooleanUnaryOperation(_BooleanOperation, ABC):
    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(operands=[toks[0][1]])


@dataclass(frozen=True)
class _BooleanBinaryOperation(_BooleanOperation, ABC):
    @classmethod
    def _create_from_tokens(cls, toks: ParseResults):
        return cls(operands=toks[0][0::2])


@dataclass(frozen=True)
class BooleanNotOperation(_BooleanUnaryOperation):
    def evaluate(self, recipe: Recipe) -> bool:
        return not self.operands[0].evaluate(recipe)

    def __invert__(self):
        return self.operands[0]


@dataclass(frozen=True)
class BooleanAndOperation(_BooleanBinaryOperation):
    def evaluate(self, recipe: Recipe) -> bool:
        return all(oper.evaluate(recipe) for oper in self.operands)

    @classmethod
    def _create_from_implicit_tokens(cls, toks: ParseResults):
        return cls(operands=list(toks[0]))

    @_value_error_to_not_implemented
    def __and__(self, other):
        return BooleanAndOperation(operands=[*self.operands, self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __rand__(self, other):
        return BooleanAndOperation(operands=[self._to_filter_element(other), *self.operands])


@dataclass(frozen=True)
class BooleanOrOperation(_BooleanBinaryOperation):
    def evaluate(self, recipe: Recipe) -> bool:
        return any(oper.evaluate(recipe) for oper in self.operands)

    @_value_error_to_not_implemented
    def __or__(self, other):
        return BooleanOrOperation(operands=[*self.operands, self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __ror__(self, other):
        return BooleanOrOperation(operands=[self._to_filter_element(other), *self.operands])


@dataclass(frozen=True)
class BooleanXorOperation(_BooleanBinaryOperation):
    OPERATOR = "xor"

    def evaluate(self, recipe: Recipe) -> bool:
        return reduce(lambda left, right: left ^ right, [oper.evaluate(recipe) for oper in self.operands])

    @_value_error_to_not_implemented
    def __xor__(self, other):
        return BooleanXorOperation(operands=[*self.operands, self._to_filter_element(other)])

    @_value_error_to_not_implemented
    def __rxor__(self, other):
        return BooleanXorOperation(operands=[self._to_filter_element(other), *self.operands])


def _normalize_str(text: str):
    """Normalizes a string for evaluation of filter strings"""
    return unicodedata.normalize("NFKD", text.strip().casefold())


class FilterParser:
    """Allows parsing filter strings into ASTs"""
    filter_expression_parser: ParserElement

    def __init__(self):
        self.filter_expression_parser = self._create_parser()

    def parse_filter_string(self, filter_string: str) -> _FilterElement:
        """
        Parses a filter string into an evaluateable AST.

        >>> fp = FilterParser()
        >>> fp.parse_filter_string('vegan and summer')
        BooleanAndOperation(operands=[AnyFilterTerm(filter_string=FuzzyFilterString(string='vegan')), AnyFilterTerm(filter_string=FuzzyFilterString(string='summer'))])

        :param filter_string: A string representing a recipe filter
        :return: filter_element: A filter ast, subclass of FilterElement
        :raises ParseBaseException: If string is not a valid filter
        :raises re.error: If a regular expression used in a filter term is not syntactically correct
        """
        return self.filter_expression_parser.parseString(filter_string, parseAll=True)[0]

    @staticmethod
    def _create_parser() -> ParserElement:
        # operators in the format later used by infixNotation
        operator_list = [
            (None, 2, opAssoc.LEFT, BooleanAndOperation._create_from_implicit_tokens),
            (CaselessKeyword('not') | "~" | "!", 1, opAssoc.RIGHT, BooleanNotOperation._create_from_tokens),
            (CaselessKeyword('and') | "&", 2, opAssoc.LEFT, BooleanAndOperation._create_from_tokens),
            (CaselessKeyword('xor') | "^", 2, opAssoc.LEFT, BooleanXorOperation._create_from_tokens),
            (CaselessKeyword('or') | "|", 2, opAssoc.LEFT, BooleanOrOperation._create_from_tokens),
        ]

        # terms (atoms) that will be combined with the boolean operators
        term_list = [
            (CaselessKeyword('tag'), TagFilterTerm._create_from_tokens),
            (CaselessKeyword('ingr'), IngredientFilterTerm._create_from_tokens),
            (CaselessKeyword('unit'), UnitFilterTerm._create_from_tokens),
            (None, AnyFilterTerm._create_from_tokens),
        ]

        # extract keywords that can
        operator_expressions = [om[0] for om in operator_list if om[0] is not None]
        term_expressions = [tm[0] for tm in term_list if tm[0] is not None]
        reserved_expressions = operator_expressions + term_expressions

        # quoted string indicates exact macthc
        quoted_filter_string = (QuotedString('"', escChar='\\') | QuotedString("'", escChar='\\')).setResultsName('string')
        # quoted_filter_string.setDebug(True)
        quoted_filter_string.setName("quoted_filter_string")
        quoted_filter_string.setParseAction(ExactFilterString._create_from_tokens)

        # not quoted string is inexact match, can't contain whitespace or be an operator
        unquoted_filter_string = ~MatchFirst(reserved_expressions) + Regex(r'[^\s\(\)]+', flags=re.U).setResultsName('string')
        # unquoted_filter_string.setDebug(True)
        unquoted_filter_string.setName("unquoted_filter_string")
        unquoted_filter_string.setParseAction(FuzzyFilterString._create_from_tokens)

        # regular expressions aren't parsed in the grammar but delegated to python re.compile in the parser action
        regex_filter_string = QuotedString('/', escChar='\\')
        regex_filter_string.setName("regex_filter_string")
        regex_filter_string.setParseAction(RegexFilterString._create_from_tokens)

        # unquoted_filter_string must be last, so that initial quotes are handled correctly
        filter_string = regex_filter_string | quoted_filter_string | unquoted_filter_string
        filter_string.setParseAction(lambda toks: toks[0])

        filter_terms = []
        for prefix_expression, term_action in term_list:
            if prefix_expression is not None:
                filter_term = Combine(prefix_expression + ':' + filter_string.setResultsName("filter_string"))
                filter_term.setName("filter_term_"+str(prefix_expression.match))
            else:
                filter_term = filter_string.setResultsName("filter_string")
                filter_term.setName("filter_term_None")
            # filter_term.setDebug(True)
            filter_term.addParseAction(term_action)
            filter_terms.append(filter_term)
        filter_term = MatchFirst(filter_terms)
        filter_expr = infixNotation(filter_term, operator_list)

        return filter_expr


# DSL
@dataclass(frozen=True)
class _FilterTermBuilder:
    term: Optional[Type[FilterTerm]] = None

    @property
    def fuz(self) -> '_FilterStringBuilder':
        """Create a filter where the given string is matched fuzzily"""
        return _FilterStringBuilder(term=self.term, filter_string=FuzzyFilterString)

    @property
    def re(self) -> '_FilterStringBuilder':
        """Create a filter where the given string is interpreted as a regular expression"""
        return _FilterStringBuilder(term=self.term, filter_string=lambda s: RegexFilterString(re.compile(s)))

    @property
    def ex(self) -> '_FilterStringBuilder':
        """Create a filter where the given string must match exactly"""
        return _FilterStringBuilder(term=self.term, filter_string=ExactFilterString)

    def __call__(self, string):
        """Create a filter where the given string is matched fuzzily"""
        term = self.term or AnyFilterTerm
        return term(FuzzyFilterString(string))


@dataclass(frozen=True)
class _FilterStringBuilder:
    term: Optional[Type[FilterTerm]]
    filter_string: Optional[Callable[[str], _FilterString]] = None

    def __call__(self, string):
        """Create a filter that matches any supported fields and the given string is matched fuzzily"""
        term = self.term or AnyFilterTerm
        filter_string = self.filter_string or FuzzyFilterString
        return term(filter_string(string))


@dataclass(frozen=True)
class FilterBuilder(_FilterTermBuilder):
    """
    Creates filter expressions in code

    Allows using a language similar to the expressions parsed by `recipemd.filter.FilterParser` directly in python code.
    This makes filters usable as an embedded domain specific language.

    A preconfigured Filter builder is available as `f` in this module:

    >>> f('Cheese')
    AnyFilterTerm(filter_string=FuzzyFilterString(string='Cheese'))

    >>> f.ingr('Cheese')
    IngredientFilterTerm(filter_string=FuzzyFilterString(string='Cheese'))

    >>> f.re('Cheese|Bacon')
    AnyFilterTerm(filter_string=RegexFilterString(regex=re.compile('Cheese|Bacon')))

    >>> f.tag.ex('vegan')
    TagFilterTerm(filter_string=ExactFilterString(string='vegan'))


    Filters can be combined with logical operators "&" (and), "|" (or), "^" (xor), and "~" (not):

    >>> f('Cheese') | f('Bacon')
    BooleanOrOperation(operands=[AnyFilterTerm(filter_string=FuzzyFilterString(string='Cheese')), AnyFilterTerm(filter_string=FuzzyFilterString(string='Bacon'))])
    """
    @property
    def any(self) -> _FilterTermBuilder:
        """Create a filter that matches any supported fields"""
        return _FilterTermBuilder(term=AnyFilterTerm)

    @property
    def tag(self) -> _FilterTermBuilder:
        """Create a filter that matches tags"""
        return _FilterTermBuilder(term=TagFilterTerm)

    @property
    def ingr(self) -> _FilterTermBuilder:
        """Create a filter that matches ingredient names"""
        return _FilterTermBuilder(term=IngredientFilterTerm)

    @property
    def unit(self) -> _FilterTermBuilder:
        """Create a filter that matches units in ingredients or in yields"""
        return _FilterTermBuilder(term=UnitFilterTerm)

    def __call__(self, string) -> AnyFilterTerm:
        return AnyFilterTerm(FuzzyFilterString(string))


#: A preconfigured :class:`FilterBuilder` for convenient use
f = FilterBuilder()

# hack to make sphinx_autodoc_typehints happy
object.__setattr__(f, '__qualname__', 'f')



import inspect
import re
from decimal import Decimal

import pytest
from pyparsing import ParseException

from recipemd.data import Recipe, Ingredient, Amount, IngredientGroup
from recipemd.filter import FilterParser, f, BooleanOrOperation, AnyFilterTerm, FuzzyFilterString, BooleanAndOperation, \
    RegexFilterString, BooleanXorOperation, BooleanNotOperation


@pytest.fixture(scope="session")
def filter_parser() -> FilterParser:
    return FilterParser()


@pytest.mark.parametrize("filter_string, expected_ast", [
    ("a b",  f("a") & f("b")),
    ("'a' 'b'", f.ex("a") & f.ex("b")),
    ("/a+/ 'b'", f.re("a+") & f.ex("b")),
    ("tag:a", f.tag("a")),
    ("unit:b xor ingr:/c/", f.unit("b") ^ f.ingr.re("c")),
    ("not 5", ~f("5")),
    ("a tag:b", f("a") & f.tag("b")),
    ("a and b", f("a") & f("b")),
    ("a and ~b?", f("a") & ~f("b?")),
    ("a & ~b?", f("a") & ~f("b?")),
    ("ingr:/^a+$/ ^ $b", f.ingr.re("^a+$") ^ f("$b")),
    ("a and b or c", (f("a") & f("b")) | f("c")),
    ("(a and b) or c", (f("a") & f("b")) | f("c")),
    ("a and (b or c)", f("a") & (f("b") | f("c"))),
    ("!/[a-c]+/", ~f.re("[a-c]+")),

    # unquoted filter term may not start with location specifier
    ("tag:", ParseException),
    # containing a location specifier is okay
    ("btag:", f("btag:")),
    # * is not a valid regex
    ("/*/", re.error),
])
def test_parse_filter_string(filter_parser, filter_string, expected_ast):
    if inspect.isclass(expected_ast) and issubclass(expected_ast, BaseException):
        with pytest.raises(expected_ast):
            filter_parser.parse_filter_string(filter_string)
    else:
        assert filter_parser.parse_filter_string(filter_string) == expected_ast


@pytest.mark.parametrize("builder_ast, expected_ast", [
    (
        "a" & f.any("b"),
        BooleanAndOperation(operands=[AnyFilterTerm(FuzzyFilterString("a")), AnyFilterTerm(FuzzyFilterString("b"))])
    ),
    (
        re.compile("a") | f("b"),
        BooleanOrOperation(operands=[AnyFilterTerm(RegexFilterString(re.compile("a"))), AnyFilterTerm(FuzzyFilterString("b"))])
    ),
    (
        ~f("a"),
        BooleanNotOperation(operands=[AnyFilterTerm(FuzzyFilterString("a"))])
    ),
    (
        ~(~f("a")),
        AnyFilterTerm(FuzzyFilterString("a"))
    ),
    (
        "a" & (f("b") & "c") & "d",
        BooleanAndOperation(operands=[
            AnyFilterTerm(FuzzyFilterString("a")), AnyFilterTerm(FuzzyFilterString("b")),
            AnyFilterTerm(FuzzyFilterString("c")), AnyFilterTerm(FuzzyFilterString("d"))
        ])

    ),
    (
        "a" | (f("b") | "c") | "d",
        BooleanOrOperation(operands=[
            AnyFilterTerm(FuzzyFilterString("a")), AnyFilterTerm(FuzzyFilterString("b")),
            AnyFilterTerm(FuzzyFilterString("c")), AnyFilterTerm(FuzzyFilterString("d"))
        ])

    ),
    (
        "a" ^ ("b" ^ f.fuz("c")) ^ FuzzyFilterString("d"),
        BooleanXorOperation(operands=[
            AnyFilterTerm(FuzzyFilterString("a")), AnyFilterTerm(FuzzyFilterString("b")),
            AnyFilterTerm(FuzzyFilterString("c")), AnyFilterTerm(FuzzyFilterString("d"))
        ])
    ),

])
def test_filter_builder(builder_ast, expected_ast):
    assert builder_ast == expected_ast


def test_filter_builder_invalid_operands():
    with pytest.raises(TypeError):
        f("a") | 5


@pytest.fixture(scope="session")
def recipe() -> Recipe:
    return Recipe(
        title="Test",
        tags=["vegetarian", "flavorful", "tag with spaces"],
        yields=[Amount(factor=Decimal("1"), unit="serving"), Amount(factor=Decimal(0.4), unit="kg")],
        ingredients=[
            Ingredient(amount=Amount(factor=Decimal('5')), name='Eggs'),
            Ingredient(amount=Amount(factor=Decimal('200'), unit='g'), name='Butter'),
            IngredientGroup(title='Group', children=[
                Ingredient(amount=Amount(factor=Decimal('2'), unit='cloves'), name='Garlic'),
                IngredientGroup(title='Subgroup', children=[
                    Ingredient(name='Onions'),
                ]),
            ]),
            Ingredient(name='Salt')
        ],
    )


@pytest.mark.parametrize("filter_ast, result", [
    (f.tag("vegetarian"), True),
    (f.tag("veg"), True),
    (f.tag.ex("vegetarian"), True),
    (f.tag.ex("veg"), False),
    (f.tag.re("flavou?rful"), True),
    (f.tag.re("veg"), True),
    (f.tag.re("^veg$"), False),
    (f.unit.ex("g"), True),
    (f.unit.ex("l"), False),
    (f.ingr("Eggs"), True),
    (f.ingr("Ham"), False),
    (f.ingr.ex("Subgroup"), False),
    (f.ex("tag with spaces"), True),

    (~f("Ham"), True),
    (f("Eggs") | f("Ham"), True),
    (f("Cheese") | f("Ham"), False),
    (f("Eggs") & f("Salt"), True),
    (f("Eggs") & f("Ham"), False),
    (f("Eggs") ^ f("Ham"), True),
    (f("Eggs") ^ f("Salt"), False),
    (f("Cheese") ^ f("Ham"), False),
])
def test_evaluate(recipe, filter_ast, result):
    assert filter_ast.evaluate(recipe) == result



import glob
import os
import textwrap
from dataclasses import replace
from decimal import Decimal

import pytest

from recipemd.data import Recipe, Amount, Ingredient, IngredientGroup, multiply_recipe, RecipeParser, RecipeSerializer, \
    get_recipe_with_yield


@pytest.fixture(scope="session")
def parser():
    return RecipeParser()


@pytest.fixture(scope="session")
def serializer():
    return RecipeSerializer()


def test_amount():
    with pytest.raises(TypeError) as excinfo:
        Amount()

    assert excinfo.value.args[0] == "Factor and unit may not both be None"


def test_recipe_get_leaf_ingredients():
    recipe = Recipe(
        title="Test",
        ingredients=[
            Ingredient(amount=Amount(factor=Decimal('5')), name='Eggs'),
            Ingredient(amount=Amount(factor=Decimal('200'), unit='g'), name='Butter'),
            IngredientGroup(title='Group', children=[
                Ingredient(amount=Amount(factor=Decimal('2'), unit='cloves'), name='Garlic'),
                IngredientGroup(title='Subgroup', children=[
                    Ingredient(name='Onions'),
                ]),
            ]),
            Ingredient(name='Salt')
        ],
    )

    leaf_ingredients = list(recipe.leaf_ingredients)
    assert len(leaf_ingredients) == 5
    assert leaf_ingredients[0].name == 'Eggs'
    assert leaf_ingredients[4].name == 'Salt'


class TestRecipeParser:
    @pytest.mark.parametrize(
        "testcase_file",
        glob.glob(os.path.join(os.path.dirname(__file__), '..', 'testcases', '*.md')),
    )
    def test_parse(self, parser, testcase_file):
        if testcase_file.endswith('.invalid.md'):
            with pytest.raises(BaseException):
                with open(testcase_file, 'r', encoding='UTF-8') as f:
                    parser.parse(f.read())
        else:
            expected_result_file = os.path.splitext(testcase_file)[0] + '.json'
            with open(expected_result_file, 'r', encoding='UTF-8') as f:
                expected_result = Recipe.from_json(f.read())
            with open(testcase_file, 'r', encoding='UTF-8') as f:
                actual_result = parser.parse(f.read())
            assert actual_result == expected_result

    def test_parse_amount(self, parser):
        assert parser.parse_amount("2") == Amount(factor=Decimal('2'))
        assert parser.parse_amount("5 g") == Amount(factor=Decimal('5'), unit='g')
        assert parser.parse_amount("5 1/4 ml") == Amount(factor=Decimal('5.25'), unit='ml')
        assert parser.parse_amount("1/4 l") == Amount(factor=Decimal('0.25'), unit='l')
        assert parser.parse_amount("-5") == Amount(factor=Decimal('-5'))
        assert parser.parse_amount("3.2") == Amount(factor=Decimal('3.2'))
        assert parser.parse_amount("3,2") == Amount(factor=Decimal('3.2'))
        assert parser.parse_amount("1 ½ cloves") == Amount(factor=Decimal('1.5'), unit='cloves')
        assert parser.parse_amount("½ pieces") == Amount(factor=Decimal('.5'), unit='pieces')
        assert parser.parse_amount('') is None


class TestRecipeSerializer:
    def test_serialize(self, serializer):
        testcase_folder = os.path.join(os.path.dirname(__file__), '..', 'testcases')
        with open(os.path.join(testcase_folder, 'recipe.md'), 'r', encoding='UTF-8') as f:
            expected_result = f.read()
        with open(os.path.join(testcase_folder, 'recipe.json'), 'r', encoding='UTF-8') as f:
            recipe = Recipe.from_json(f.read())
        actual_result = serializer.serialize(recipe)
        assert actual_result == expected_result

    def test_serialize_amount(self, serializer):
        assert serializer._serialize_amount(Amount(factor=Decimal('5.000'))) == '5'
        assert serializer._serialize_amount(Amount(factor=Decimal('1')/Decimal('3')), rounding=2) == '0.33'
        assert serializer._serialize_amount(Amount(factor=Decimal('1')/Decimal('3')), rounding=4) == '0.3333'


def test_multiply_recipe():
    recipe = Recipe(
        title="Test",
        yields=[Amount(factor=Decimal('5'), unit="servings")],
        ingredients=[
            Ingredient(amount=Amount(factor=Decimal('5')), name='Eggs'),
            Ingredient(amount=Amount(factor=Decimal('200'), unit='g'), name='Butter'),
            IngredientGroup(title='Group', children=[
                Ingredient(amount=Amount(factor=Decimal('2'), unit='cloves'), name='Garlic')
            ]),
            Ingredient(name='Salt')
        ],
    )

    result = multiply_recipe(recipe, Decimal(2))

    assert result.yields[0].factor == Decimal('10')
    assert result.ingredients[0].amount.factor == Decimal('10')
    assert result.ingredients[1].amount.factor == Decimal('400')
    assert result.ingredients[2].children[0].amount.factor == Decimal('4')
    assert result.ingredients[3].amount is None


def test_get_recipe_with_yield():
    recipe = Recipe(
        title="Test",
        yields=[Amount(factor=Decimal('2'), unit="servings")],
        ingredients=[
            Ingredient(amount=Amount(factor=Decimal('5')), name='Eggs'),
        ],
    )

    result = get_recipe_with_yield(recipe, Amount(factor=Decimal('4'), unit='servings'))
    assert result.yields[0] == Amount(factor=Decimal('4'), unit='servings')
    assert result.ingredients[0].amount == Amount(factor=Decimal('10'))

    # interpreted as "4 recipes", that is multiply by 4
    result_unitless = get_recipe_with_yield(recipe, Amount(factor=Decimal('4')))
    assert result_unitless.yields[0] == Amount(factor=Decimal('8'), unit='servings')
    assert result_unitless.ingredients[0].amount == Amount(factor=Decimal('20'))

    # if recipe has unitless yield, it is preferred to the above interpretation
    recipe_with_unitless_yield = replace(recipe, yields=[Amount(factor=Decimal('4'))])
    result_unitless_from_unitless_yield = get_recipe_with_yield(recipe_with_unitless_yield, Amount(factor=Decimal('4')))
    assert result_unitless_from_unitless_yield.yields[0] == Amount(factor=Decimal('4'))
    assert result_unitless_from_unitless_yield.ingredients[0].amount == Amount(factor=Decimal('5'))

    # try with unit not in recipe yields
    with pytest.raises(StopIteration):
        get_recipe_with_yield(recipe, Amount(factor=Decimal('500'), unit='ml'))



# PYTHON_ARGCOMPLETE_OK
"""
Implements :ref:`cli_recipemd`
"""

import argparse
import decimal
import json
import os
import re
import sys
import urllib.parse
import urllib.request
from dataclasses import replace
from typing import List, Union, Dict, Optional

import argcomplete
from argcomplete.completers import ChoicesCompleter, FilesCompleter
from yarl import URL

import recipemd
from recipemd.data import RecipeParser, RecipeSerializer, multiply_recipe, Ingredient, get_recipe_with_yield, \
    IngredientGroup, Recipe, Amount

__all__ = ['main']


def main():
    # completions
    argcomplete.autocomplete(parser)

    # parse args
    args = parser.parse_args()

    # initialize
    rp = RecipeParser()
    rs = RecipeSerializer()

    # read and parse recipe
    src = args.file.read()
    r = rp.parse(src)

    # scale recipe
    r = _process_scaling(r, args)

    # base url for late use
    base_url = URL(f'file://{os.path.abspath(args.file.name)}')

    # export linked recipes
    if args.export_links:
        _export_links(r, args, base_url, rp, rs)
        return

    # flatten recipe
    if args.flatten:
        r = _get_flattened_recipe(r, base_url=base_url, parser=rp)

    # create output depending on arguments
    print(_create_recipe_output(r, rs, args))


def _yield_completer(prefix, action, parser, parsed_args):
    try:
        src = parsed_args.file.read()
        r = RecipeParser().parse(src)

        parsed_yield = RecipeParser.parse_amount(prefix)
        if parsed_yield is None or parsed_yield.factor is None:
            return [RecipeSerializer._serialize_amount(a) for a in r.yields]

        return [RecipeSerializer._serialize_amount(Amount(parsed_yield.factor, a.unit)) for a in r.yields
                if parsed_yield.unit is None or (a.unit is not None and a.unit.startswith(parsed_yield.unit))]
    except Exception as e:
        print(e)
        return []


def _process_scaling(r, args):
    """Returns recipes scaled according to --multiply or --yield"""
    if args.required_yield is not None:
        required_yield = RecipeParser.parse_amount(args.required_yield)
        if required_yield is None or required_yield.factor is None:
            print(f'Given yield is not valid', file=sys.stderr)
            exit(1)
        try:
            r = get_recipe_with_yield(r, required_yield)
        except StopIteration:
            print(f'Recipe "{r.title}" does not specify a yield in the unit "{required_yield.unit}". The '
                  f'following units can be used: ' + ", ".join(f'"{y.unit}"' for y in r.yields), file=sys.stderr)
            exit(1)
    elif args.multiply is not None:
        multiply = RecipeParser.parse_amount(args.multiply)
        if multiply is None or multiply.factor is None:
            print(f'Given multiplier is not valid', file=sys.stderr)
            exit(1)
        if multiply.unit is not None:
            print(f'A recipe can only be multiplied with a unitless amount', file=sys.stderr)
            exit(1)
        r = multiply_recipe(r, multiply.factor)
    return r


def _export_links(r, args, base_url, parser, serializer):
    if type(args.export_links) == bool:
        folder = args.file.name.rsplit('.', 1)[0]
    else:
        folder = args.export_links
    os.makedirs(folder, exist_ok=True)
    print(f'Writing to {folder}', file=sys.stderr)
    link_ingredients, ingr_to_recipe = _get_linked_recipes(r, base_url=base_url, parser=parser, flatten=True)
    for ingredient in link_ingredients:
        try:
            recipe = ingr_to_recipe[ingredient]
        except KeyError:
            continue
        url = base_url.join(URL(ingredient.link))
        filename = os.path.join(folder, url.parts[-1])
        with open(filename, 'w') as f:
            print(f'Created {filename} for "{recipe.title}"', file=sys.stderr)
            f.write(_create_recipe_output(recipe, serializer, args))


def _create_recipe_output(recipe, serializer, args):
    """Serializes a recipes according to formatting options"""
    if args.title:
        return recipe.title
    elif args.ingredients:
        return "\n".join(_ingredient_to_string(ingr, rounding=args.round) for ingr in recipe.leaf_ingredients)
    elif args.json:
        return recipe.to_json()
    else:
        return serializer.serialize(recipe, rounding=args.round)


def _get_flattened_recipe(recipe: Recipe, *, base_url: URL, parser: RecipeParser) -> Recipe:
    """Creates a new recipe with linked recipes recursively flattened"""
    link_ingredients, ingr_to_recipe = _get_linked_recipes(recipe, base_url=base_url, parser=parser)

    # recipes that contain no links need not be processed
    if not ingr_to_recipe:
        return recipe

    # ingredients
    recipe = replace(recipe, ingredients=_create_flattened_substituted_ingredients(recipe.ingredients, ingr_to_recipe))

    # instructions
    instruction_sections = []
    for ingredient in link_ingredients:
        try:
            link_recipe = ingr_to_recipe[ingredient]
        except KeyError:
            pass
        else:
            if link_recipe.instructions:
                instruction_sections.append((_link_ingredient_title(ingredient, link_recipe), link_recipe.instructions, False))

    if recipe.instructions:
        instruction_sections.append((recipe.title, recipe.instructions, True))

    instructions = []
    if len(instruction_sections) == 1 and instruction_sections[0][2]:
        instructions.append(instruction_sections[0][1])
    else:
        for heading, body, is_main_instructions in instruction_sections:
            # find headings (see https://spec.commonmark.org/0.29/#atx-heading) and increase level by one
            # note that only up to level 6 is allowed, so we will do 5 -> 6 but not 6 -> 7
            new_body = re.sub(r'^( {0,3})(#{1,5}.*)$', r'\1#\2', body, flags=re.MULTILINE)
            instructions.append(f'## {heading}\n\n{new_body}')

    recipe = replace(recipe, instructions='\n\n'.join(instructions))

    return recipe


def _get_linked_recipes(recipe: Recipe, *, base_url: URL, parser: RecipeParser, flatten: bool = True):
    """Gets all ingredients that have a link and a dict of the ingredient id() to recipe instance"""
    link_ingredients = [i for i in recipe.leaf_ingredients if i.link is not None]
    ingr_to_recipe = dict()
    for ingredient in link_ingredients:
        try:
            ingr_to_recipe[ingredient] = _get_linked_recipe(ingredient, base_url=base_url, parser=parser,
                                                            flatten=flatten)
        except Exception as e:
            print(f'{e}: {e.__cause__}', file=sys.stderr)
    return link_ingredients, ingr_to_recipe


def _get_linked_recipe(ingredient: Ingredient, *, base_url: URL, parser: RecipeParser, flatten: bool = True) -> Recipe:
    url = base_url.join(URL(ingredient.link))
    try:
        with urllib.request.urlopen(str(url)) as req:
            encoding = req.info().get_content_charset() or 'UTF-8'
            src = req.read().decode(encoding)
    except Exception as e:
        raise RuntimeError(f'''Couldn't find linked recipe for ingredient "{ingredient.name}"''') from e

    try:
        link_recipe = parser.parse(src)
    except Exception as e:
        raise RuntimeError(f'''Couldn't parse linked recipe for ingredient "{ingredient.name}"''') from e

    if flatten:
        link_recipe = _get_flattened_recipe(link_recipe, base_url=url, parser=parser)

    if ingredient.amount:
        link_recipe = get_recipe_with_yield(link_recipe, ingredient.amount)

    return link_recipe


def _create_flattened_substituted_ingredients(ingredients: List[Union[Ingredient, IngredientGroup]],
                                              ingr_to_recipe: Dict[Ingredient, Recipe]) \
        -> List[Union[Ingredient, IngredientGroup]]:
    result_ingredients = []
    result_groups = []
    for ingr in ingredients:
        if isinstance(ingr, IngredientGroup):
            new_group = replace(ingr, children=_create_flattened_substituted_ingredients(ingr.children, ingr_to_recipe))
            result_groups.append(new_group)
        elif ingr in ingr_to_recipe and ingr_to_recipe[ingr] is not None:
            link_recipe = ingr_to_recipe[ingr]
            new_group = IngredientGroup(title=_link_ingredient_title(ingr, link_recipe), children=link_recipe.ingredients)
            result_groups.append(new_group)
        else:
            result_ingredients.append(ingr)

    # groups must come after ingredients, see https://github.com/tstehr/RecipeMD/issues/6
    return result_ingredients + result_groups


def _link_ingredient_title(ingr: Ingredient, link_recipe: Recipe) -> str:
    if ingr.name == link_recipe.title:
        title = f'[{ingr.name}]({ingr.link})'
    else:
        title = f'[{ingr.name}: {link_recipe.title}]({ingr.link})'
    return title


def _ingredient_to_string(ingr: Ingredient, *, rounding: Optional[int]=None) -> str:
    if ingr.amount is not None:
        return f'{RecipeSerializer._serialize_amount(ingr.amount, rounding=rounding)} {ingr.name}'
    return ingr.name


# parser is on module level for sphinx-autoprogram
parser = argparse.ArgumentParser(description='Read and process recipemd recipes')

parser.add_argument(
    'file', type=argparse.FileType('r', encoding='UTF-8'), help='A recipemd file'
).completer = FilesCompleter(allowednames='*.md')

parser.add_argument('-v', '--version', action='version', version=f"%(prog)s ({recipemd.__version__})")

display_parser = parser.add_mutually_exclusive_group()
display_parser.add_argument('-t', '--title', action='store_true', help='Display recipe title')
display_parser.add_argument('-i', '--ingredients', action='store_true', help='Display recipe ingredients')
display_parser.add_argument('-j', '--json', action='store_true', help='Display recipe as JSON')

parser.add_argument(
    '-r', '--round', type=lambda s: None if s.lower() == 'no' else int(s), metavar='n', default=2,
    help='Round amount to n digits after decimal point. Default is "2", use "no" to disable rounding.'
).completer = ChoicesCompleter(('no', *range(0, decimal.getcontext().prec + 1)))

scale_parser = parser.add_mutually_exclusive_group()
scale_parser.add_argument('-m', '--multiply', type=str, help='Multiply recipe by N', metavar='N')
scale_parser.add_argument(
    '-y', '--yield', type=str, help='Scale the recipe for yield Y, e.g. "5 servings"',
    metavar='Y', dest='required_yield'
).completer = _yield_completer

flatten_parser = parser.add_mutually_exclusive_group()
flatten_parser.add_argument(
    '-f', '--flatten', action='store_true',
    help='Flatten ingredients and instructions of linked recipes into main recipe'
)
flatten_parser.add_argument(
    '--export-links', type=str, metavar='DIR', nargs='?', default=False, const=True,
    help='Export flattened linked recipes as required for the main recipe to DIR (DIR defaults to recipe file name)'
)


if __name__ == "__main__":
    main()



# PYTHON_ARGCOMPLETE_OK
"""
Implements :ref:`cli_recipemd_find`
"""

import argparse
import collections
import glob
import itertools
import os
import re
import shutil
import sys
import unicodedata
from math import floor, ceil
from typing import Callable, Iterable

import argcomplete
import pyparsing
from argcomplete import FilesCompleter

import recipemd
from recipemd.data import RecipeParser, Recipe
from recipemd.filter import _FilterElement, FilterParser

__all__ = ['main']


def main():
    # completions
    argcomplete.autocomplete(parser)

    args = parser.parse_args()
    args.func(args)


def list_recipes(args):
    print_result([path for recipe, path in get_filtered_recipes(args)], args.output_multicol)


def list_tags(args):
    list_elements(args, lambda recipe: recipe.tags)


def list_ingredients(args):
    list_elements(args, lambda recipe: [ingr.name for ingr in recipe.leaf_ingredients if ingr.name is not None])


def get_units(recipe: Recipe) -> Iterable[str]:
    ingredient_units = (ingr.amount.unit for ingr in recipe.leaf_ingredients if
                        ingr.amount is not None and ingr.amount.unit is not None)
    yield_units = (yield_.unit for yield_ in recipe.yields if yield_.unit is not None)
    return itertools.chain(ingredient_units, yield_units)


def list_units(args):
    list_elements(args, lambda recipe: get_units(recipe))


def list_elements(args, extractor: Callable[[Recipe], Iterable[str]]):
    counter = collections.Counter()

    for recipe, path in get_filtered_recipes(args):
        counter.update(extractor(recipe))

    if args.count:
        result = list(counter.items())
        result.sort(key=lambda pair: pair[1], reverse=True)
        max_count_length = max(len(str(c)) for c in counter.values())
        result = [f'{count:>{max_count_length}} {tag}' for tag, count in result]
    else:
        result = list(counter)
        result.sort(key=lambda s: s.casefold())

    print_result(result, args.output_multicol)


def get_filtered_recipes(args):
    rp = RecipeParser()
    result = []
    for path in glob.glob(os.path.join(args.folder, '**/*.md'), recursive=True):
        try:
            with open(path, 'r', encoding='UTF-8') as file:
                recipe = rp.parse(file.read())
            if args.expression is None or args.expression.evaluate(recipe):
                result.append((recipe, os.path.relpath(path, args.folder)))
        except Exception as e:
            if not args.no_messages:
                print(f"An error occurred, skipping {os.path.relpath(path, args.folder)}: {e.args[0]}", file=sys.stderr)
    return result


def print_result(items, output_multicol):
    if output_multicol is None:
        if os.isatty(sys.stdout.fileno()):
            output_multicol = 'columns'
        else:
            output_multicol = 'no'

    if output_multicol == 'columns':
        print_columns(items)
    elif output_multicol == 'rows':
        print_columns(items, transpose=True)
    else:
        print("\n".join(items))


def print_columns(items, transpose=True):
    if not items:
        return

    # normalize items so decomposed unicode chars don't break lines
    items = [unicodedata.normalize('NFKC', item) for item in items]
    max_item_width = max(len(item) for item in items)
    column_width = max_item_width + 2
    line_width, _ = shutil.get_terminal_size((80, 20))

    # calculate column and line count
    column_count = floor(line_width / column_width)
    if column_count == 0:
        column_count = 1
    row_count = ceil(len(items) / column_count)

    if transpose:
        matrix = [items[i::row_count] for i in range(0, row_count)]
    else:
        matrix = [items[i:i+column_count] for i in range(0, len(items), column_count)]

    for row in matrix:
        padded_row = [val.ljust(column_width) for val in row[0:-1]]
        padded_row.append(row[-1])
        print("".join(padded_row))


def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f'"{path}" is not a valid folder')


def create_filter_expr(filter_string) -> _FilterElement:
    try:
        return FilterParser().parse_filter_string(filter_string)
    except pyparsing.ParseBaseException as e:
        raise argparse.ArgumentTypeError(f'"{filter_string}" is not a valid filter: {e}')
    except re.error as e:
        raise argparse.ArgumentTypeError(f'"{filter_string}" contains the regular expression "{e.pattern}": {e}')


# parser is on module level for sphinx-autoprogram
parser = argparse.ArgumentParser(description='Find recipes, ingredients and units by filter expression')

parser.add_argument('-v', '--version', action='version', version=f"%(prog)s ({recipemd.__version__})")

parser.add_argument(
    '-e', '--expression', type=create_filter_expr,
    help='Filter expression. Expects a boolean string, e.g. "cake and vegan or ingr:cheese"'
)
parser.add_argument('-s', '--no-messages', action='store_true', default=False, help='suppress error messages')

matrix_parser = parser.add_mutually_exclusive_group()
matrix_parser.add_argument(
    '-1', dest='output_multicol', action='store_const', const='no',
    help='Force output to be one entry per line. This is the default when output is not to a terminal.'
)
matrix_parser.add_argument(
    '-C', dest='output_multicol', action='store_const', const='columns',
    help='Force multi-column output; this is the default when output is to a terminal.'
)
matrix_parser.add_argument(
    '-x', dest='output_multicol', action='store_const', const='rows',
    help='The same as -C, except that the multi-column output is produced with entries sorted across, rather than '
         'down, the columns.'
)

subparsers = parser.add_subparsers(metavar="action", required=True)

# recipes
parser_recipes = subparsers.add_parser('recipes', help='list recipe paths')
parser_recipes.set_defaults(func=list_recipes)

parser_recipes.add_argument(
    'folder', type=dir_path, nargs='?', default='.', help='path to a folder containing recipemd files. Works '
                                                          'recursively for all *.md files.'
    # very unlikely file extension so completer only returns folders
).completer = FilesCompleter(allowednames="*.7CA0B927-3B02-48EA-97A9-CB557E061992")

# list tags
parser_tags = subparsers.add_parser('tags', help="list used tags")
parser_tags.set_defaults(func=list_tags)

parser_tags.add_argument('-c', '--count', action='store_true', help="count number of uses per tag")
parser_tags.add_argument(
    'folder', type=dir_path, nargs='?', default='.', help='path to a folder containing recipemd files. Works '
                                                          'recursively for all *.md files.'
    # very unlikely file extension so completer only returns folders
).completer = FilesCompleter(allowednames="*.7CA0B927-3B02-48EA-97A9-CB557E061992")

# list ingredients
parser_ingredients = subparsers.add_parser('ingredients', help="list used ingredients")
parser_ingredients.set_defaults(func=list_ingredients)

parser_ingredients.add_argument('-c', '--count', action='store_true', help="count number of uses per ingredient")
parser_ingredients.add_argument(
    'folder', type=dir_path, nargs='?', default='.', help='path to a folder containing recipemd files. Works '
                                                          'recursively for all *.md files.'
    # very unlikely file extension so completer only returns folders
).completer = FilesCompleter(allowednames="*.7CA0B927-3B02-48EA-97A9-CB557E061992")

# list units
parser_units = subparsers.add_parser('units', help="list used units")
parser_units.set_defaults(func=list_units)

parser_units.add_argument('-c', '--count', action='store_true', help="count number of uses per unit")
parser_units.add_argument(
    'folder', type=dir_path, nargs='?', default='.', help='path to a folder containing recipemd files. Works '
                                                          'recursively for all *.md files.'
    # very unlikely file extension so completer only returns folders
).completer = FilesCompleter(allowednames="*.7CA0B927-3B02-48EA-97A9-CB557E061992")

# TODO edit
# parser_edit = subparsers.add_parser('edit', help='edit tags')


if __name__ == "__main__":
    main()



import setuptools
from recipemd import __version__

tests_require = [
    'pytest',
    'pytest-cov',
]

docs_require = [
    'Sphinx',
    'm2r',
    'sphinxcontrib.fulltoc',
    'sphinxcontrib.autoprogram',
    'sphinx_autodoc_typehints',
    'sphinxcontrib.apidoc',
    'sphinx-autobuild'
]

release_requires = [
    'twine',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="recipemd",
    version=__version__,
    author="Tilman Stehr",
    author_email="tilman@tilman.ninja",
    description="Markdown recipe manager, reference implementation of RecipeMD",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://recipemd.org",
    packages=setuptools.find_packages(),
    python_requires='>=3.7,<4',
    install_requires=[
        'dataclasses-json>=0.3.0,<0.4.0',
        'yarl~=1.3.0',
        # has broken before and is still unstable, so pin exact version
        'commonmarkextensions==0.0.5',
        # commonmark version needs to match the version required by commonmarkextensions
        'commonmark>=0.8.0,<=0.8.1',
        'argcomplete~=1.10.0',
        'pyparsing~=2.4.2',
    ],
    extras_require={
        'tests': tests_require,
        'docs': docs_require,
        'release': release_requires,
        'dev': tests_require+docs_require+release_requires
    },
    entry_points={
        'console_scripts': [
            'recipemd=recipemd.cli.main:main',
            'recipemd-find=recipemd.cli.find:main',
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
        "Operating System :: OS Independent",
    ],
)



"""
Defines the RecipeMD data structures, provides parser, serializer and recipe scaling functions.
"""
#from __future__ import annotations

import re
import unicodedata
from dataclasses import dataclass, field, replace
from decimal import Decimal
from typing import List, Optional, Union, Dict, Generator

import commonmark
from commonmark.node import Node
from commonmark_extensions.plaintext import CommonMarkToCommonMarkRenderer
from dataclasses_json import dataclass_json, config

__all__ = ['RecipeParser', 'RecipeSerializer', 'multiply_recipe', 'get_recipe_with_yield',
           'Recipe', 'Ingredient', 'IngredientGroup', 'Amount']


def _decode_ingredient_element_list(ingrs: List[Dict]):
    """Helper for json decoding List[Union[Ingredient, IngredientGroup]]"""
    return [_decode_ingredient_element(el) for el in ingrs]


def _decode_ingredient_element(ingr_el_dict: Dict):
    """
    Helper for json decoding Union[Ingredient, IngredientGroup]

    dataclasses-json can't figure out Union[Ingredient, IngredientGroup], so we use this duck typing decoder to help
    """
    if "children" in ingr_el_dict:
        return IngredientGroup.from_dict(ingr_el_dict)
    return Ingredient.from_dict(ingr_el_dict)


@dataclass_json
@dataclass(frozen=True)
class IngredientGroup:
    title: Optional[str] = None
    children: List[Union['Ingredient', 'IngredientGroup']] = field(
        default_factory=list,
        metadata=config(decoder=_decode_ingredient_element_list),
    )


@dataclass_json
@dataclass(frozen=True)
class Amount:
    factor: Optional[Decimal] = field(
        default=None,
        # decoder is workaround for https://github.com/lidatong/dataclasses-json/issues/137
        metadata=config(decoder=lambda val: Decimal(val) if val is not None else None)
    )
    unit: Optional[str] = None

    def __post_init__(self):
        if self.factor is None and self.unit is None:
            raise TypeError(f"Factor and unit may not both be None")


@dataclass_json
@dataclass(frozen=True)
class Ingredient:
    name: str
    amount: Optional[Amount] = None
    link: Optional[str] = None


@dataclass_json
@dataclass(frozen=True)
class Recipe:
    title: str
    description: Optional[str] = None
    yields: List[Amount] = field(default_factory=list)
    tags: List[str] = field(default_factory=list)
    ingredients: List[Union[Ingredient, IngredientGroup]] = field(
        default_factory=list,
        metadata=config(decoder=_decode_ingredient_element_list),
    )
    instructions: Optional[str] = None

    @property
    def leaf_ingredients(self) -> Generator['Ingredient', None, None]:
        yield from self._get_leaf_ingredients(self.ingredients)

    def _get_leaf_ingredients(self, ingredients: List[Union[Ingredient, IngredientGroup]]):
        for ingr in ingredients:
            if isinstance(ingr, IngredientGroup):
                yield from self._get_leaf_ingredients(ingr.children)
            else:
                yield ingr


class RecipeSerializer:
    def serialize(self, recipe: Recipe, *, rounding: Optional[int] = None) -> str:
        rep = ""
        rep += f'# {recipe.title}\n\n'
        if recipe.description is not None:
            rep += f'{recipe.description}\n\n'
        if len(recipe.tags) > 0:
            rep += f'*{", ".join(recipe.tags)}*\n\n'
        if len(recipe.yields) > 0:
            rep += f'**{", ".join(self._serialize_amount(a, rounding=rounding) for a in recipe.yields)}**\n\n'
        rep += f'---\n\n'
        rep += ("\n".join(self._serialize_ingredient(g, 2, rounding=rounding) for g in recipe.ingredients)).strip()
        if recipe.instructions is not None:
            rep += f'\n\n---\n\n'
            rep += recipe.instructions
        return rep

    def _serialize_ingredient(self, ingredient, level, *, rounding: Optional[int] = None):
        if isinstance(ingredient, IngredientGroup):
            return f'\n{"#" * level} {ingredient.title}\n\n'\
                   + "\n".join(self._serialize_ingredient(i, level+1, rounding=rounding) for i in ingredient.children)
        else:
            if ingredient.amount is not None:
                return f'- *{self._serialize_amount(ingredient.amount, rounding=rounding)}* {self._serialize_ingredient_text(ingredient)}'
            return f'- {self._serialize_ingredient_text(ingredient)}'

    @staticmethod
    def _serialize_ingredient_text(ingredient):
        if ingredient.link:
            return f'[{ingredient.name}]({ingredient.link})'
        return ingredient.name

    @staticmethod
    def _serialize_amount(amount, *, rounding: Optional[int] = None):
        if amount.factor is not None and amount.unit is not None:
            return f'{RecipeSerializer._normalize_factor(amount.factor, rounding=rounding)} {amount.unit}'
        if amount.factor is not None:
            return f'{RecipeSerializer._normalize_factor(amount.factor, rounding=rounding)}'
        if amount.unit is not None:
            return f'{amount.unit}'

    @staticmethod
    def _normalize_factor(factor: Decimal, *, rounding: Optional[int]=None):
        if rounding is not None:
            factor = round(factor, rounding)
        # remove trailing zeros (https://docs.python.org/3/library/decimal.html#decimal-faq)
        factor = factor.quantize(Decimal(1)) if factor == factor.to_integral() else factor.normalize()
        return factor


class RecipeParser:
    _list_split = re.compile("(?<!\d),|,(?!\d)")

    src: Optional[str]
    recipe: Optional[Recipe]
    parents: List[Node]
    current: Node

    def __init__(self):
        self.src = None
        self.recipe = None
        self.parents = []
        self.current = None

    def parse(self, src: str) -> Recipe:
        self.src = src
        self.recipe = Recipe(title=None)

        parser = commonmark.Parser()
        ast = parser.parse(src)

        self.current = ast.first_child

        self._parse_title()
        self._parse_description()
        self._parse_tags_and_yields()

        if self.current is not None and self.current.t == 'thematic_break':
            self._next_node()
        else:
            # TODO this divider is required, but we might just continue anyways?
            raise RuntimeError(f"Invalid, expected divider before ingredient list, got {self.current and self.current.t} instead")

        self._parse_ingredients()

        if self.current is not None:
            if self.current.t == 'thematic_break':
                self._next_node()
            else:
                # TODO this divider is required, but we might just continue anyways?
                raise RuntimeError(f"Invalid, expected divider before ingredient list, got {self.current and self.current.t} instead")

        self.recipe = replace(self.recipe, instructions=self._get_source_until())

        return self.recipe

    def _parse_title(self):
        if self.current is not None and self.current.t == 'heading' and self.current.level == 1:
            self.recipe = replace(self.recipe, title=self._get_current_node_children_source())
            self._next_node()
        else:
            # TODO title is required according to spec, maybe the parser might be more foregiving?
            raise RuntimeError(f"Invalid, title (heading with level 1) required, got "
                               f"{self.current and self.current.t if not self.current or self.current.t != 'heading' else f'level {self.current.level }'} "
                               f"instead")

    def _parse_description(self):
        result = self._get_source_until(lambda n: n.t != 'thematic_break' and not self._is_tags(n)
                                        and not self._is_yields(n))
        self.recipe = replace(self.recipe, description=result)

    def _parse_tags_and_yields(self):
        while self.current is not None and (self._is_tags(self.current) or self._is_yields(self.current)):
            if self._is_tags(self.current):
                self._enter_node()
                tags_text = self._get_current_node_children_source()
                self._exit_node()
                self.recipe = replace(self.recipe, tags=[t.strip() for t in self._list_split.split(tags_text)])
            else:
                self._enter_node()
                yields_text = self._get_current_node_children_source()
                self._exit_node()
                self.recipe = replace(self.recipe, yields=[self.parse_amount(t.strip()) for t in self._list_split.split(yields_text)])
            self._next_node()

    def _parse_ingredients(self):
        while self.current is not None and (self.current.t == 'heading' or self.current.t == 'list'):
            if self.current.t == 'heading':
                self._parse_ingredient_groups(self.recipe.ingredients, parent_level=-1)
            elif self.current is not None and self.current.t == 'list':
                self._parse_ingredient_list_node(self.recipe.ingredients)

    def _parse_ingredient_groups(self, ingredients, parent_level):
        while self.current is not None and self.current.t == 'heading':
            level = self.current.level
            if level <= parent_level:
                return

            group = IngredientGroup(title=self._get_current_node_children_source())
            self._next_node()

            if self.current is not None and self.current.t == 'list':
                self._parse_ingredient_list_node(group.children)

            ingredients.append(group)

            self._parse_ingredient_groups(group.children, parent_level=level)

    def _parse_ingredient_list_node(self, ingredients):
        self._enter_node()
        while self.current is not None:
            ingredients.append(self._parse_ingredient())
            self._next_node()
        self._exit_node()
        self._next_node()

    def _parse_ingredient(self):
        self._enter_node()

        amount = None
        first_paragraph = None

        can_have_link = True
        has_link = False
        link_destination = None
        link_text = None

        if self.current is not None and self.current.t == 'paragraph':
            # enter paragraph
            self._enter_node()

            if self.current.t == 'emph':
                # enter emph
                self._enter_node()
                amount = self.parse_amount(self._get_node_source(self.current))
                # leave emph
                self._exit_node()
                # parse rest of first paragraph
                self._next_node()

            while self.current is not None:
                if first_paragraph is None:
                    first_paragraph = ""
                node_source = self._get_node_source(self.current)

                if can_have_link:
                    # to have a link, the first paragraph may only contain space and one link
                    if not has_link and self.current.t == 'link':
                        link_destination = self.current.destination
                        link_text = self._get_current_node_children_source()
                        has_link = True
                    elif not node_source.isspace():
                        can_have_link = False

                first_paragraph += node_source
                self._next_node()

            # leave first paragraph
            self._exit_node()
            self._next_node()

        following_paragraphs = self._get_source_until()

        if following_paragraphs is not None and not following_paragraphs.isspace():
            can_have_link = False

        if can_have_link and has_link:
            name = link_text
        elif first_paragraph is not None and following_paragraphs is not None:
            name = first_paragraph + "\n\n" + following_paragraphs
        elif first_paragraph is not None:
            name = first_paragraph
        elif following_paragraphs is not None:
            name = following_paragraphs
        else:
            raise RuntimeError("No ingredient name!")
        name = name.strip()

        self._exit_node()

        return Ingredient(name=name, amount=amount, link=link_destination)

    _value_formats = [
        # improper fraction (1 1/2)
        (r'(\d+)\s+(\d+)\s*/\s*(\d+)', lambda match: Decimal(match.group(2)) + (Decimal(match.group(3)) / Decimal(match.group(4))), 3),
        # improper fraction with unicode vulgar fraction (1 ½)
        (r'(\d+)\s+([\u00BC-\u00BE\u2150-\u215E])', lambda match: Decimal(match.group(2)) + Decimal(unicodedata.numeric(match.group(3))), 2),
        # proper fraction (5/6)
        (r'(\d+)\s*/\s*(\d+)', lambda match: Decimal(match.group(2)) / Decimal(match.group(3)), 2),
        # proper fraction with unicode vulgar fraction (⅚)
        (r'([\u00BC-\u00BE\u2150-\u215E])', lambda match: Decimal(unicodedata.numeric(match.group(2))), 1),
        # decimal (5,4 or 5.6)
        (r'(\d*)[.,](\d+)', lambda match: Decimal(match.group(2) + '.' + match.group(3)), 2),
        # integer (4)
        (r'(\d+)', lambda match: Decimal(match.group(2)), 1)
    ]

    @staticmethod
    def parse_amount(amount_str: str) -> Amount:
        # iterate over different value format
        for regexp, factor_function, group_count in RecipeParser._value_formats:
            match = re.match(r'^\s*(-?)\s*' + regexp + r'(.*)$', amount_str)
            if match:
                factor = factor_function(match)
                if match.group(1) == '-':
                    factor = -1 * factor
                unit = match.group(group_count + 2).strip()
                return Amount(factor, unit or None)
            
        unit = amount_str.strip()
        return Amount(None, unit) if unit else None

    def _is_tags(self, ast_node: Node):
        return ast_node.t == 'paragraph' and ast_node.first_child.t == 'emph' \
               and ast_node.first_child == ast_node.last_child

    def _is_yields(self, ast_node: Node):
        return ast_node.t == 'paragraph' and ast_node.first_child.t == 'strong' \
               and ast_node.first_child == ast_node.last_child

    def _next_node(self):
        self.current = self.current.nxt

    def _enter_node(self):
        self.parents.append(self.current)
        self.current = self.current.first_child

    def _exit_node(self):
        self.current = self.parents.pop()

    def _get_source_until(self, predicate=None):
        if self.current is None:
            return None

        start = self.current.sourcepos[0]
        end = None

        while self.current is not None and (predicate is None or predicate(self.current)):
            end = self.current.sourcepos[1]
            self._next_node()

        if end is not None:
            return self._get_sourcepos_source((start, end))

        return None

    def _get_node_source(self, ast_node):
        if ast_node.sourcepos is not None:
            return self._get_sourcepos_source(ast_node.sourcepos)
        if ast_node.literal is not None:
            return ast_node.literal
        return CommonMarkToCommonMarkRenderer().render(ast_node)

    def _get_current_node_children_source(self):
        source = ""
        if self.current.first_child is not None:
            self._enter_node()
            while self.current is not None:
                source += self._get_node_source(self.current)
                self._next_node()
            self._exit_node()
        return source

    def _get_sourcepos_source(self, sourcepos):
        start_line = sourcepos[0][0] - 1
        end_line = sourcepos[1][0]

        first_line_start_offset = sourcepos[0][1] - 1
        last_line_end_offset = sourcepos[1][1]

        lines = self.src.splitlines()[start_line:end_line]
        lines[0] = lines[0][first_line_start_offset:]
        lines[-1] = lines[-1][:last_line_end_offset]

        return "\n".join(lines)


def multiply_recipe(recipe: Recipe, multiplier: Decimal) -> Recipe:
    """
    Multiplies a recipe by the given multiplier

    Creates a new recipe where the factor of yield and ingredient is changed according to the multiplier


    """
    recipe = replace(recipe, yields=[replace(y, factor=y.factor * multiplier) for y in recipe.yields])
    recipe = replace(recipe, ingredients=_multiply_ingredients(recipe.ingredients, multiplier))
    return recipe


def get_recipe_with_yield(recipe: Recipe, required_yield: Amount) -> Recipe:
    """
    Scale the given recipe to a required yield

    Raises `StopIteration` if no yield with a matching unit can be found.
    """
    matching_recipe_yield = next((y for y in recipe.yields if y.unit == required_yield.unit), None)
    if matching_recipe_yield is None:
        # no unit in required amount is interpreted as "one recipe"
        if required_yield.unit is None:
            matching_recipe_yield = Amount(Decimal(1))
        else:
            raise StopIteration
    recipe = multiply_recipe(recipe, required_yield.factor / matching_recipe_yield.factor)
    return recipe


def _multiply_ingredients(ingredients: List[Union[Ingredient, IngredientGroup]], multiplier: Decimal):
    return [_multiply_ingredient(ingr, multiplier) for ingr in ingredients]


def _multiply_ingredient(ingr: Union[Ingredient, IngredientGroup], multiplier: Decimal) -> Union[Ingredient, IngredientGroup]:
    if hasattr(ingr, 'amount') and ingr.amount is not None and ingr.amount.factor is not None:
        return replace(ingr, amount=replace(ingr.amount, factor=ingr.amount.factor*multiplier))
    if hasattr(ingr, 'children'):
        return replace(ingr, children=_multiply_ingredients(ingr.children, multiplier))
    return ingr
